public class LinkedPositionalList<E> implements PositionalList<E> {
  protected int size;            	// Number of elements in the list
  protected Node<E> header, trailer;	// Special sentinels

  /** Constructor that creates an empty list; O(1) time */
  public LinkedPositionalList() {
    size = 0;
    header = new Node<E>(null, null, null);	// create header
    trailer = new Node<E>(null, header, null);	// create trailer
    header.setNext(trailer);	// make header and trailer point to each other
  }
  /** Checks if position is valid for this list and converts it to
    *  Node if it is valid; O(1) time */
  protected Node<E> validatePosition(Position<E> p)
    throws InvalidPositionException {
    if (p == null)
      throw new InvalidPositionException
	("Null position passed to NodeList");
    if (p == header)
	throw new InvalidPositionException
	  ("The header node is not a valid position");
    if (p == trailer)
	throw new InvalidPositionException
	  ("The trailer node is not a valid position");
    try {
      Node<E> temp = (Node<E>) p;
      if ((temp.getPrev() == null) || (temp.getNext() == null))
	throw new InvalidPositionException
	  ("Position does not belong to a valid NodeList");
      return temp;
    } catch (ClassCastException e) {
      throw new InvalidPositionException
	("Position is of wrong type for this list");
    }
  }
  /** Returns the number of elements in the list;  O(1) time */
  public int size() { return size; }
  /** Returns whether the list is empty;  O(1) time  */
  public boolean isEmpty() { return (size == 0); }
  /** Returns the first position in the list; O(1) time */
  public Position<E> first()
      throws EmptyListException {
    if (isEmpty())
      throw new EmptyListException("List is empty");
    return header.getNext();
  }
  /** Returns the last position in the list; O(1) time */
  public Position<E> last()
      throws EmptyListException {
    if (isEmpty())
      throw new EmptyListException("List is empty");
    return trailer.getPrev();
  }
  /** Returns the position after the given one; O(1) time */
  public Position<E> after(Position<E> p)
      throws InvalidPositionException, BoundaryViolationException {
    Node<E> v = validatePosition(p);
    Node<E> next = v.getNext();
    if (next == trailer)
      throw new BoundaryViolationException
	("Cannot advance past the end of the list");
    return next;
  }
  /** Returns the position before the given one; O(1) time */
  public Position<E> before(Position<E> p)
      throws InvalidPositionException, BoundaryViolationException {
    Node<E> v = validatePosition(p);
    Node<E> prev = v.getPrev();
    if (prev == header)
      throw new BoundaryViolationException
	("Cannot advance past the beginning of the list");
    return prev;
  }
  /** Insert the given element after the given position, returning
    * the new position; O(1) time  */
  public void addAfter(Position<E> p, E element) 
      throws InvalidPositionException {			// 
    Node<E> v = validatePosition(p);
    size++;
    Node<E> newNode = new Node<E>(element, v, v.getNext());
    v.getNext().setPrev(newNode);
    v.setNext(newNode);
  }
  /** Insert the given element before the given position, returning
    * the new position; O(1) time  */
  public void addBefore(Position<E> p, E element) 
      throws InvalidPositionException {			// 
    Node<E> v = validatePosition(p);
    size++;
    Node<E> newNode = new Node<E>(element, v.getPrev(), v);
    v.getPrev().setNext(newNode);
    v.setPrev(newNode);
  }
  /** Insert the given element at the beginning of the list, returning
    * the new position; O(1) time  */
  public void addFirst(E element) {
    size++;
    Node<E> newNode = new Node<E>(element, header, header.getNext());
    header.getNext().setPrev(newNode);
    header.setNext(newNode);
  }
  /** Insert the given element at the end of the list, returning
    * the new position; O(1) time  */
  public void addLast(E element) {
    size++;
    Node<E> newNode = new Node<E>(element, trailer.getPrev(), trailer);
    trailer.getPrev().setNext(newNode);
    trailer.setPrev(newNode);
  }
  /**Remove the given position from the list; O(1) time */
  public E remove(Position<E> p)
      throws InvalidPositionException {
    Node<E> v = validatePosition(p);
    size--;
    Node<E> vPrev = v.getPrev();
    Node<E> vNext = v.getNext();
    vPrev.setNext(vNext);
    vNext.setPrev(vPrev);
    E vElem = v.getElement();
    // unlink the position from the list and make it invalid
    v.setNext(null);
    v.setPrev(null);
    return vElem;
  }
  /** Replace the element at the given position with the new element
    * and return the old element; O(1) time  */
  public E set(Position<E> p, E element)
      throws InvalidPositionException {
    Node<E> v = validatePosition(p);
    E oldElt = v.getElement();
    v.setElement(element);
    return oldElt;
  }
}
